# Copyright 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import vassert, scenario, expected, test

from reahl.web_dev.fixtures import WebFixture

from reahl.component.exceptions import ProgrammerError
from reahl.web.fw import Bookmark, Url
from reahl.web.ui import A, Div

from reahl.web.bootstrap.navs import Nav, PillLayout, TabLayout, DropdownMenu, DropdownMenuLayout



@test(WebFixture)
def navs(fixture):
    """A Nav is a Menu with css classes for styling by Bootstrap."""

    bookmarks = [Bookmark('', '/one', 'One'),
                 Bookmark('', '/two', 'Two')]
    menu = Nav(fixture.view).with_bookmarks(bookmarks)
    
    # A nav is an ul.nav
    vassert( menu.html_representation.tag_name == 'ul' )
    vassert( 'nav' in menu.html_representation.get_attribute('class') )

    # Containing a li for each menu item
    [one, two] = menu.html_representation.children

    for item, expected_href, expected_description in [(one, '/one', 'One'),
                                                      (two, '/two', 'Two')]:
        vassert( item.tag_name == 'li' )
        vassert( item.get_attribute('class') == 'nav-item' )

        [a] = item.children
        vassert( a.get_attribute('href') == expected_href )
        vassert( a.children[0].value ==  expected_description )
        vassert( a.get_attribute('class') == 'nav-link' )


class VisualFeedbackScenarios(WebFixture):
    @scenario
    def disabled(self):
        """The mouse cursor is shown as no-access on disabled items."""
        def not_allowed():
            return False
        self.menu_item_with_state = A(self.view, Url('/another_url'), write_check=not_allowed)
        self.state_indicator_class = 'disabled'

    @scenario
    def active(self):
        """The currently active item is highlighted."""
        current_url = Url(self.request.url)
        self.menu_item_with_state = A(self.view, current_url)
        self.state_indicator_class = 'active'


@test(VisualFeedbackScenarios)
def visual_feedback_on_items(fixture):
    """The state of a MenuItem is visually indicated to a user."""
    menu = Nav(fixture.view, [])
    menu.add_a(A(fixture.view, Url('/an_url')))
    menu.add_a(fixture.menu_item_with_state)

    [defaulted_item, item_with_state] = menu.html_representation.children

    [defaulted_a] = defaulted_item.children
    [a_with_state] = item_with_state.children
    
    vassert( fixture.state_indicator_class not in defaulted_a.get_attribute('class') )
    vassert( fixture.state_indicator_class in a_with_state.get_attribute('class') )



class LayoutScenarios(WebFixture):
    @scenario
    def pill_layouts(self):
        """PillLayouts are used to make Navs MenuItems look almost like buttons."""
        self.layout_css_class = {'nav-pills'}
        self.layout = PillLayout()

    @scenario
    def stacked_pill_layouts(self):
        """Using a PillLayout, you can also make MenuItems appear stacked on top of 
           one another instead of being placed next to one another."""
        self.layout_css_class = {'nav-pills', 'nav-stacked'}
        self.layout = PillLayout(stacked=True)

    @scenario
    def tab_layouts(self):
        """TabLayouts are used to make Navs MenuItems look like a series of tabs."""
        self.layout_css_class = {'nav-tabs'}
        self.layout = TabLayout()


@test(LayoutScenarios)
def nav_layouts(fixture):
    """Navs can be laid out in different ways."""
    menu = Nav(fixture.view)

    vassert( not fixture.layout_css_class.issubset(menu.html_representation.attributes['class'].value) )
    menu.use_layout(fixture.layout)
    vassert( fixture.layout_css_class.issubset(menu.html_representation.attributes['class'].value) )


class DifferentLayoutTypes(WebFixture):
    @scenario
    def pills(self):
        self.layout_type = PillLayout

    @scenario
    def tabs(self):
        self.layout_type = TabLayout


@test(DifferentLayoutTypes)
def justified_items(fixture):
    """Both a PillLayout or TabLayout can be set to make the MenuItems of
       their Nav fill the width of the parent, with the text of each item centered."""

    menu = Nav(fixture.view).use_layout(fixture.layout_type())
    vassert( 'nav-justified' not in menu.html_representation.get_attribute('class') )
    
    menu = Nav(fixture.view).use_layout(fixture.layout_type(justified=True))
    vassert( 'nav-justified' in menu.html_representation.get_attribute('class') )


@test(WebFixture)
def pill_layouts_cannot_mix_justified_and_stacked(fixture):
    """A PillLayout cannot be both stacked and justified at the same time."""

    with expected(ProgrammerError):
        PillLayout(stacked=True, justified=True)


@test(WebFixture)
def dropdown_menus(fixture):
    """You can add a DropdownMenu as a dropdown inside a Nav."""
    menu = Nav(fixture.view)
    sub_menu = DropdownMenu(fixture.view)
    sub_menu.add_a(A(fixture.view, Url('/an/url'), description='sub menu item'))
    menu.add_dropdown('Dropdown title', sub_menu)

    [item] = menu.html_representation.children
    
    vassert( item.tag_name == 'li' )
    vassert( 'dropdown' in item.get_attribute('class') )
    
    [toggle, added_sub_menu] = item.children
    vassert( 'dropdown-toggle' in toggle.get_attribute('class') )
    vassert( 'dropdown' in toggle.get_attribute('data-toggle') )
    vassert( '-' in toggle.get_attribute('data-target') )
    vassert( 'caret' in toggle.children[1].get_attribute('class') )

    title_text = toggle.children[0].value
    vassert( title_text == 'Dropdown title' )

    vassert( added_sub_menu is sub_menu )
    vassert( 'dropdown-menu' in added_sub_menu.html_representation.get_attribute('class').split() )
    vassert( isinstance(added_sub_menu.html_representation, Div) )

    [dropdown_item] = added_sub_menu.html_representation.children
    vassert( isinstance(dropdown_item, A) )
    vassert( 'dropdown-item' in dropdown_item.get_attribute('class').split() )
    

@test(WebFixture)
def dropdown_menus_can_drop_up(fixture):
    """Dropdown menus can drop upwards instead of downwards."""
    menu = Nav(fixture.view)
    sub_menu = Nav(fixture.view)
    menu.add_dropdown('Dropdown title', sub_menu, drop_up=True)

    [item] = menu.html_representation.children
    
    vassert( item.tag_name == 'li' )
    vassert( 'dropup' in item.get_attribute('class') )


@test(WebFixture)
def dropdown_menus_right_align(fixture):
    """Dropdown menus can be aligned to the bottom right of their toggle, instead of the default (left)."""

    defaulted_sub_menu = DropdownMenu(fixture.view).use_layout(DropdownMenuLayout())
    vassert( 'dropdown-menu-right' not in defaulted_sub_menu.html_representation.get_attribute('class') )

    right_aligned_sub_menu = DropdownMenu(fixture.view).use_layout(DropdownMenuLayout(align_right=True))
    vassert( 'dropdown-menu-right' in right_aligned_sub_menu.html_representation.get_attribute('class') )








# Copyright 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
.. versionadded:: 3.2

Utilities to deal with layout.


"""

import six
if six.PY2:
    from collections import Mapping
else:
    from collections.abc import Mapping

import copy
from collections import OrderedDict


from reahl.component.exceptions import arg_checks, IsInstance
from reahl.web.fw import Layout
from reahl.web.ui import Div, HTML5Page, Slot, Header, Footer

class ResponsiveSize(Mapping):
    """Represents a set of relative sizes to be used for an element depending on the size of the user's device.

       This class lets one specify the size to use for different sizes of display device.

       Each kwarg of the constructor is a size and the name of the kwarg is the device class for which
       that size applies.

       Values for sizes and device classes are as defined by the underlying layout library.

       .. versionadded:: 3.2

    """
    def __init__(self, **sizes):
        self.sizes = {size_name: size_value for (size_name, size_value) in sizes.items() if size_value}

    def __getitem__(self, item):
        return self.sizes.__getitem__(item)
        
    def __iter__(self):
        return self.sizes.__iter__()
        
    def __len__(self):
        return self.sizes.__len__()
        
        
class ColumnLayout(Layout):
    """A Layout that divides an element into a number of columns.

       Each argument passed to the constructor defines a
       column. Columns are added to the element using this Layout in
       the order they are passed to the constructor. Columns can also 
       be added to the Widget later, by calling :meth:`ColumnLayout.add_column`.

       To define a column with a given :class:`ResponsiveSize`, pass a tuple of which
       the first element is the column name, and the second an
       instance of :class:`ResponsiveSize`.

       .. versionadded:: 3.2

    """
    def __init__(self, *column_definitions):
        super(ColumnLayout, self).__init__()
        self.add_slots = False
        self.columns = OrderedDict()  #: A dictionary containing the added columns, keyed by column name.
        self.column_sizes = OrderedDict()
        for column_definition in column_definitions:
            if isinstance(column_definition, tuple):
                name, size = column_definition
            else:
                name, size = column_definition, ResponsiveSize()
            self.column_sizes[name] = size

    def with_slots(self):
        """Returns a copy of this ColumnLayout which will additionally add a Slot inside each added column,
           named for that column.
        """
        copy_with_slots = copy.deepcopy(self)
        copy_with_slots.add_slots = True
        return copy_with_slots

    def customise_widget(self):
        for name, size in self.column_sizes.items():
             column = self.add_column(size)
             self.columns[name] = column
             column.append_class('column-%s' % name)
             if self.add_slots:
                 column.add_child(Slot(self.view, name))
                 
    def add_column(self, size):
        """Add an un-named column, with the given :class:`ResponsiveSize`.

           :keyword size: The sizes to use for the column.

           Returns a :class:`reahl.web.ui.Div` representing the added column.
        """
        return self.widget.add_child(Div(self.view))



class PageLayout(Layout):
    """A PageLayout creates a basic skeleton inside an :class:`reahl.web.ui.HTML5Page`, and optionally
       applies specified :class:`~reahl.web.ui.Layout`\s to parts of this skeleton.

       The skeleton consists of a :class:`~reahl.web.ui.Div` called the `document` of the page, which
       contains three sub-sections inside of it:
       
         - the `.header` -- the page header area where menus and banners go;
         - the `.contents` of the page -- the main area to which the main content will be added; and
         - the `.footer` -- the page footer where links and legal notices go.    

       :keyword document_layout: A :class:`~reahl.web.ui.Layout` that will be applied to `.document`.
       :keyword contents_layout: A :class:`~reahl.web.ui.Layout` that will be applied to `.contents`.
       :keyword header_layout: A :class:`~reahl.web.ui.Layout` that will be applied to `.header`.
       :keyword footer_layout: A :class:`~reahl.web.ui.Layout` that will be applied to `.footer`.

       .. admonition:: Styling
       
          Adds a <div id="doc"> to the <body> of the page, which contains:

           - a <header id="hd">
           - a <div id="contents">
           - a <footer id="ft">

       .. versionadded:: 3.2

    """
    @arg_checks(document_layout=IsInstance(Layout, allow_none=True), contents_layout=IsInstance(Layout, allow_none=True), 
                header_layout=IsInstance(Layout, allow_none=True), footer_layout=IsInstance(Layout, allow_none=True))
    def __init__(self, document_layout=None, contents_layout=None, header_layout=None, footer_layout=None):
        super(PageLayout, self).__init__()
        self.header = None    #: The :class:`reahl.web.ui.Header` of the page.
        self.contents = None  #: The :class:`reahl.web.ui.Div` containing the contents.
        self.footer = None    #: The :class:`reahl.web.ui.Footer` of the page.
        self.document = None  #: The :class:`reahl.web.ui.Div` containing the entire page.
        self.header_layout = header_layout   #: A :class:`reahl.web.fw.Layout` to be used for the header of the page.
        self.contents_layout = contents_layout #: A :class:`reahl.web.fw.Layout` to be used for the contents div of the page.
        self.footer_layout = footer_layout   #: A :class:`reahl.web.fw.Layout` to be used for the footer of the page.
        self.document_layout = document_layout #: A :class:`reahl.web.fw.Layout` to be used for the document of the page.

    @arg_checks(widget=IsInstance(HTML5Page))
    def apply_to_widget(self, widget):
        super(PageLayout, self).apply_to_widget(widget)

    def customise_widget(self):
        self.document = self.widget.body.add_child(Div(self.view))
        self.document.set_id('doc')
        if self.document_layout:
            self.document.use_layout(self.document_layout)
        
        self.header = self.document.add_child(Header(self.view))
        self.header.add_child(Slot(self.view, 'header'))
        self.header.set_id('hd')
        if self.header_layout:
            self.header.use_layout(self.header_layout)

        self.contents = self.document.add_child(Div(self.view))
        if self.contents_layout:
            self.contents.use_layout(self.contents_layout)

        self.contents.set_id('bd')
        self.contents.set_attribute('role', 'main')
        
        self.footer = self.document.add_child(Footer(self.view))
        self.footer.add_child(Slot(self.view, 'footer'))
        self.footer.set_id('ft')
        if self.footer_layout:
            self.footer.use_layout(self.footer_layout)




# Copyright 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

import webob

from reahl.tofu import vassert, scenario, expected, test
from reahl.stubble import stubclass

from reahl.web.fw import Widget, WebExecutionContext
from reahl.web.ui import Img
from reahl.web_dev.fixtures import WebFixture
from reahl.web.bootstrap.carousel import Carousel




class CarouselFixture(WebFixture):
    def get_main_div_for(self, carousel):
        return carousel.children[0]

    def get_indicator_list_for(self, carousel):
        ol = self.get_main_div_for(carousel).children[0]
        assert self.is_carousel_indicator(ol)
        return ol.children

    def is_carousel_indicator(self, element):
        return 'carousel-indicators' in element.get_attribute('class')

    def carousel_indicators_present(self, carousel):
        return any([self.is_carousel_indicator(child) 
                    for child in self.get_main_div_for(carousel).children])


@test(CarouselFixture)
def carousel_basics(fixture):
    """A Carousel contains all the right classes and contents to act as a Bootstrap Carousel component."""

    widget = Carousel(fixture.view, 'my_carousel_id')

    [main_div] = widget.children

    # The main div
    vassert( main_div.get_attribute('id') == 'my_carousel_id' )
    vassert( main_div.get_attribute('class') == 'carousel slide' )

    [indicator_list, carousel_inner, left_control, right_control] = main_div.children

    # Indicators
    vassert( indicator_list.get_attribute('class') == 'carousel-indicators' )

    # Inner (container of the images)
    vassert( carousel_inner.get_attribute('class') == 'carousel-inner' )
    vassert( carousel_inner.get_attribute('role') == 'listbox' )

    # Controls
    def check_control(control, position, action, label):
        vassert( control.get_attribute('class') == 'carousel-control %s' % position )
        vassert( control.get_attribute('role') == 'button' )
        vassert( control.get_attribute('data-slide') == action )

        [icon, text] = control.children
        vassert( icon.get_attribute('class') == 'icon-%s' % action )
        vassert( icon.get_attribute('aria-hidden') == 'true' )
        
        vassert( text.children[0].value == label)
        vassert( text.get_attribute('class') == 'sr-only')

    check_control(left_control, 'left', 'prev', 'Previous')
    check_control(right_control, 'right', 'next', 'Next')


@test(CarouselFixture)
def i18n(fixture):
    """User-visible labels are internationalised."""
    @stubclass(WebExecutionContext)
    class AfrikaansContext(WebExecutionContext):
        request = webob.Request.blank('/', charset='utf8')
        @property
        def interface_locale(self):
            return 'af'

    with AfrikaansContext():
        widget = Carousel(fixture.view, 'my_carousel_id')

        [main_div] = widget.children
        [indicator_list, carousel_inner, left_control, right_control] = main_div.children

        def check_control(control, label):
            [icon, text] = control.children
            vassert( text.children[0].value == label)

        check_control(left_control, 'Vorige')
        check_control(right_control, 'Volgende')



@test(CarouselFixture)
def carousel_has_options(fixture):
    """Constructor allows you to set certain customizing options"""

    carousel = Carousel(fixture.view, 'my_carousel_id', interval=1000, pause='hover', wrap=True, keyboard=True)
    main_div = fixture.get_main_div_for(carousel)

    vassert( main_div.get_attribute('data-interval') == '1000' )
    vassert( main_div.get_attribute('data-pause') == 'hover' )
    vassert( main_div.get_attribute('data-wrap') == 'true' )
    vassert( main_div.get_attribute('data-keyboard') == 'true' )



@test(CarouselFixture)
def adding_items_to_carousel(fixture):
    """Images can be added to a Carousel."""

    carousel = Carousel(fixture.view, 'my_carousel_id', show_indicators=True)
    main_div = fixture.get_main_div_for(carousel)
    [indicator_list, carousel_inner, left_control, right_control] = main_div.children

    # Initially, no items or indicators are present
    vassert( carousel_inner.children == [])
    vassert( fixture.carousel_indicators_present(carousel) )
    vassert( fixture.get_indicator_list_for(carousel) == [] )

    image = Img(fixture.view)
    added_item = carousel.add_slide(image)

    # A carousel item was added for the image
    [carousel_item] = carousel_inner.children
    vassert( carousel_item is added_item )
    vassert( 'carousel-item' in carousel_item.get_attribute('class') )

    [actual_image] = carousel_item.children
    vassert( actual_image is image )

    # An indicator was added for the item
    [indicator] = fixture.get_indicator_list_for(carousel)

    vassert( indicator.get_attribute('data-target') == '#my_carousel_id' )
    vassert( indicator.get_attribute('data-slide-to') == '0' )


@test(CarouselFixture)
def active_state_of_items(fixture):
    """The first item added is marked as active, and also its corresponding indicator."""
    
    carousel = Carousel(fixture.view, 'my_carousel_id')
    carousel.add_slide(Img(fixture.view))
    carousel.add_slide(Img(fixture.view))

    main_div = fixture.get_main_div_for(carousel)
    [indicator_list, carousel_inner, left_control, right_control] = main_div.children
    
    #only the first item is active
    [carousel_item_1, carousel_item_2] = carousel_inner.children
    vassert( carousel_item_1.get_attribute('class') == 'active carousel-item' )
    vassert( carousel_item_2.get_attribute('class') == 'carousel-item' )

    #only the first indicator is active
    [indicator_0, indicator_1] = fixture.get_indicator_list_for(carousel)

    vassert( indicator_0.get_attribute('class') == 'active' )
    vassert( not indicator_1.has_attribute('class') )
    vassert( indicator_0.get_attribute('data-slide-to') == '0' )
    vassert( indicator_1.get_attribute('data-slide-to') == '1' )


@test(CarouselFixture)
def item_indicators_are_optional(fixture):
    """With show_indicators=False, indicators are not added when adding items."""
    carousel = Carousel(fixture.view, 'my_carousel_id', show_indicators=False)

    vassert( not fixture.carousel_indicators_present(carousel) )

    carousel.add_slide(Img(fixture.view))

    #after adding the item, indicators shouldn't appear
    vassert( not fixture.carousel_indicators_present(carousel) )


@test(CarouselFixture)
def adding_items_with_captions(fixture):
    """A Widget can be supplied to be used caption for an added image."""
    carousel = Carousel(fixture.view, 'my_carousel_id')

    caption_widget = Widget(fixture.view)
    carousel_item = carousel.add_slide(Img(fixture.view), caption_widget=caption_widget)

    [image, div_containing_caption] = carousel_item.children
    vassert( div_containing_caption.get_attribute('class') == 'carousel-caption' )

    [actual_caption_widget] = div_containing_caption.children
    vassert( actual_caption_widget is caption_widget )




# Copyright 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
.. versionadded:: 3.2

Sometimes you need to display a long list of items. Displaying such a list on a single
page is not a good idea, because the page will take forever to load.

This module provides a few classes you can use to build a single View that displays
only one "pageful" of the list. You can then also include a :class:`PageMenu` -- a
menu on which a user can choose to navigate to another section (or page) of the list.



"""

from __future__ import print_function, unicode_literals, absolute_import, division
import six


from reahl.component.i18n import Translator
from reahl.component.modelinterface import exposed
from reahl.web.fw import Bookmark
from reahl.web.ui import HTMLWidget, Menu, AccessRightAttributes, ActiveStateAttributes
from reahl.web.bootstrap.ui import A, Span
from reahl.web.pager import PagedPanel, PageIndex, SequentialPageIndex, AnnualItemOrganiserProtocol, AnnualPageIndex


_ = Translator('reahl-web')


class PageMenu(HTMLWidget):
    """An Menu, which lists the pages of items that can be navigated by a user. If there are
       many pages, only a small subset is shown, with controls allowing the user to browse to
       the wanted page number and choose it.

       :param view: (See :class:`reahl.web.fw.Widget`)
       :param css_id: (See :class:`reahl.web.ui.HTMLElement`)
       :param page_index: The :class:`PageIndex` whose pages are displayed by this PageMenu.
       :param paged_panel: The :class:`PagedPanel` in which the contents of a page is displayed.

    """
    def __init__(self, view, css_id, page_index, paged_panel):
        self.page_index = page_index
        self.paged_panel = paged_panel
        super(PageMenu, self).__init__(view)

        self.menu = self.add_child(Menu(view, add_reahl_styling=False))
        self.create_items(self.menu)
        self.set_html_representation(self.menu)
        self.set_id(css_id)
        self.append_class('pagination')

        self.enable_refresh(self.query_fields.start_page_number)

    def add_styling_to_menu_item(self, item):
        item.a.append_class('page-link')
        item.widget.append_class('page-item')

    def create_items(self, menu):
        links = []

        self.add_bordering_link_for(menu, '←', 'First', 1,
                                   not self.page_index.has_previous_page)
        self.add_bordering_link_for(menu, '«', 'Prev', self.page_index.previous_page.number, 
                                   not self.page_index.has_previous_page)

        for page in self.page_index.pages_in_range:
            bookmark = self.paged_panel.get_bookmark(page_number=page.number, description=page.description)
            bookmark.query_arguments['start_page_number'] = self.page_index.start_page_number
            link = A.from_bookmark(self.view, bookmark)
            item = menu.add_a(link)
            item.widget.add_attribute_source(ActiveStateAttributes(item))
            self.add_styling_to_menu_item(item)
            if self.page_index.current_page_number == page.number:
                item.set_active()

        self.add_bordering_link_for(menu, '»', 'Next', self.page_index.next_page.number, 
                                   not self.page_index.has_next_page)
        self.add_bordering_link_for(menu, '→', 'Last', self.page_index.last_page.number, 
                                   not self.page_index.has_next_page)


    def add_bordering_link_for(self, menu, short_description, long_description, start_page_number, disabled):
        link = A.from_bookmark(self.view, self.get_bookmark(start_page_number=start_page_number, 
                                                            disabled=disabled))

        link.add_child(Span(self.view, text=short_description)).set_attribute('aria-hidden', 'true');
        link.add_child(Span(self.view, text=long_description)).append_class('sr-only');
        link.set_attribute('aria-label', long_description);
        link.set_active(not disabled)
        item = menu.add_a(link)
        item.widget.add_attribute_source(AccessRightAttributes(link))
        self.add_styling_to_menu_item(item)
        
    def get_bookmark(self, start_page_number=1, disabled=False):
        bookmark = Bookmark.for_widget(None,
                                       query_arguments={'start_page_number': start_page_number,
                                                        'current_page_number': self.page_index.current_page_number},
                                       write_check=lambda: not disabled).on_view(self.view)
        return bookmark
    

    @exposed
    def query_fields(self, fields):
        fields.start_page_number = self.page_index.fields.start_page_number
        fields.current_page_number = self.paged_panel.query_fields.current_page_number

    @property
    def jquery_selector(self):
        return '"ul.pagination"'

    def get_js(self, context=None):
        js = ['$(%s).bootstrappagemenu({});' % self.jquery_selector]
        return super(PageMenu, self).get_js(context=context) + js 









# Copyright 2013, 2014, 2015 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
import re
import json
from webob import Response

from reahl.stubble import stubclass, CallMonitor
from nose.tools import istest
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert, expected

from reahl.webdev.tools import Browser
from reahl.component.exceptions import DomainException
from reahl.web.fw import CheckedRemoteMethod
from reahl.web.fw import JsonResult
from reahl.web.fw import MethodResult
from reahl.web.fw import RemoteMethod, RegenerateMethodResult
from reahl.web.fw import Widget
from reahl.web.fw import WidgetResult
from reahl.component.modelinterface import Field, IntegerField
from reahl.component.py3compat import ascii_as_bytes_or_str
from reahl.web_dev.fixtures import  WebFixture

class RemoteMethodFixture(WebFixture):
    def new_widget_factory(self, remote_method=None):
        remote_method = remote_method or self.remote_method
        
        @stubclass(Widget)
        class WidgetWithRemoteMethod(Widget):
            def __init__(self, view):
                super(WidgetWithRemoteMethod, self).__init__(view)
                view.add_resource(remote_method)
                
        return WidgetWithRemoteMethod.factory()

    def new_wsgi_app(self, remote_method=None):
        remote_method = remote_method or self.remote_method
        slot_definitions = {'main': self.new_widget_factory(remote_method=remote_method)}
        return super(RemoteMethodFixture, self).new_wsgi_app(view_slots=slot_definitions)
            
        
@istest
class RemoteMethodTests(object):
    @test(WebFixture)
    def remote_methods(self, fixture):
        """A RemoteMethod is a SubResource representing a method on the server side which can be invoked via POSTing to an URL."""

        def callable_object():
            return 'value returned from method'

        encoding = 'koi8_r'  # Deliberate
        remote_method = RemoteMethod('amethod', callable_object, MethodResult(mime_type='ttext/hhtml', encoding=encoding))
    
        @stubclass(Widget)
        class WidgetWithRemoteMethod(Widget):
            def __init__(self, view):
                super(WidgetWithRemoteMethod, self).__init__(view)
                view.add_resource(remote_method)

        wsgi_app = fixture.new_wsgi_app(view_slots={'main': WidgetWithRemoteMethod.factory()})
        browser = Browser(wsgi_app)
        
        # By default you cannot GET, since the method is not immutable
        browser.open('/_amethod_method', status=405)
        
        # POSTing to the URL, returns the result of the method
        browser.post('/_amethod_method', {})
        vassert( browser.raw_html == 'value returned from method' )
        vassert( browser.last_response.charset == encoding)
        vassert( browser.last_response.content_type == 'ttext/hhtml' )
    
    @test(RemoteMethodFixture)
    def exception_handling(self, fixture):
        """The RemoteMethod sends back the six.text_type() of an exception raised for the specified exception class."""
        
        def fail():
            raise Exception('I failed')
        remote_method = RemoteMethod('amethod', fail, MethodResult(catch_exception=Exception))

        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)

        browser.post('/_amethod_method', {})
        vassert( browser.raw_html == 'I failed' )

    @test(RemoteMethodFixture)
    def immutable_remote_methods(self, fixture):
        """A RemoteMethod that is immutable is accessible via GET (instead of POST)."""

        def callable_object():
            return 'value returned from method'
        remote_method = RemoteMethod('amethod', callable_object, MethodResult(), immutable=True)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        # GET, since the method is immutable
        browser.open('/_amethod_method')
        vassert( browser.raw_html == 'value returned from method' )
        
        # POSTing to the URL, is not supported
        browser.post('/_amethod_method', {}, status=405)
    
    
    class ArgumentScenarios(RemoteMethodFixture):
        @scenario
        def get(self):
            self.immutable = True
        @scenario
        def post(self):
            self.immutable = False
            
    @test(ArgumentScenarios)
    def arguments_to_remote_methods(self, fixture):
        """A RemoteMethod can get arguments from a query string or submitted form values, depending on the scenario."""

        def callable_object(**kwargs):
            fixture.method_kwargs = kwargs 
            return ''
        remote_method = RemoteMethod('amethod', callable_object, MethodResult(), immutable=fixture.immutable)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        kwargs_sent = {'a':'AAA', 'b':'BBB'}
        if fixture.immutable:
            browser.open('/_amethod_method?a=AAA&b=BBB')
        else:
            browser.post('/_amethod_method', kwargs_sent)
        vassert( fixture.method_kwargs == kwargs_sent )

    @test(ArgumentScenarios)
    def checked_arguments(self, fixture):
        """A CheckedRemoteMethod checks and marshalls its parameters using Fields."""

        def callable_object(anint=None, astring=None):
            fixture.method_kwargs = {'anint': anint, 'astring': astring}
            return ''
        remote_method = CheckedRemoteMethod('amethod', callable_object, MethodResult(), 
                                            immutable=fixture.immutable, 
                                            anint=IntegerField(),
                                            astring=Field())
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        if fixture.immutable:
            browser.open('/_amethod_method?anint=5&astring=SupercalifraGilisticexpialidocious')
        else:
            browser.post('/_amethod_method', {'anint':'5', 'astring':'SupercalifraGilisticexpialidocious'})
        vassert( fixture.method_kwargs == {'anint':5, 'astring':'SupercalifraGilisticexpialidocious'} )

    class ResultScenarios(RemoteMethodFixture):
        @scenario
        def json(self):
            self.method_result = JsonResult(IntegerField(), catch_exception=Exception)
            self.value_to_return = 1
            self.expected_response = '1'
            self.exception_response = '"exception text"'
            self.expected_charset = self.method_result.encoding
            self.expected_content_type = 'application/json'
            self.results_match = lambda x, y: x == y

        @stubclass(Widget)
        class WidgetStub(object):
            css_id = 'someid'
            def render_contents(self): return '<the widget contents>'
            def get_contents_js(self, context=None): return ['some', 'some', 'javascript']

        @scenario
        def widget(self):
            self.method_result = WidgetResult(self.WidgetStub())
            self.value_to_return = 'ignored in this case'
            self.expected_response = '<the widget contents><script type="text/javascript">javascriptsome</script>'
            self.exception_response = Exception
            self.expected_charset = self.method_result.encoding
            self.expected_content_type = 'text/html'
            self.results_match = lambda x, y: x == y

        @scenario
        def widget_as_json(self):
            self.method_result = WidgetResult(self.WidgetStub(), as_json_and_result=True)
            self.value_to_return = 'ignored in this case'
            self.expected_response = {'widget': '<the widget contents><script type="text/javascript">javascriptsome</script>',
                                      'success': True}
            self.expected_charset = self.method_result.encoding
            self.expected_content_type = 'application/json'
            def results_match(expected, actual):
                return json.loads(actual) == expected
            self.results_match = results_match
            
    @test(ResultScenarios)
    def different_kinds_of_result(self, fixture):
        """Different kinds of MethodResult can be specified for a method."""

        def callable_object():
            return fixture.value_to_return
        remote_method = RemoteMethod('amethod', callable_object, default_result=fixture.method_result)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        browser.post('/_amethod_method', {})
        vassert( fixture.results_match(fixture.expected_response, browser.raw_html) )
        vassert( browser.last_response.charset == fixture.expected_charset )
        vassert( browser.last_response.content_type == fixture.expected_content_type )

    @test(ResultScenarios.json)
    def exception_handling_for_json(self, fixture):
        """How exceptions are handled with JsonResult."""

        def fail():
            raise Exception('exception text')
        remote_method = RemoteMethod('amethod', fail, default_result=fixture.method_result)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        browser.post('/_amethod_method', {})
        vassert( browser.raw_html == fixture.exception_response )
        vassert( browser.last_response.charset == fixture.expected_charset )
        vassert( browser.last_response.content_type == fixture.expected_content_type )

    @test(ResultScenarios.widget)
    def exception_handling_for_widgets(self, fixture):
        """How exceptions are handled with WidgetResult."""

        def fail():
            raise Exception('exception text')
        remote_method = RemoteMethod('amethod', fail, default_result=fixture.method_result)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)

        with expected(Exception):
            browser.post('/_amethod_method', {})

    class RegenerateMethodResultScenarios(RemoteMethodFixture):
        method_called = 0
        def new_method_result(self):
            fixture = self
            @stubclass(MethodResult)
            class ResultThatDependsOnExecutionOfMethod(MethodResult):
                def create_response(self, return_value):
                    return Response(body='success: method was called %s times' % fixture.method_called)

                def create_exception_response(self, exception):
                    return Response(body='exception: method was called %s times' % fixture.method_called)
            return ResultThatDependsOnExecutionOfMethod(catch_exception=DomainException, replay_request=True)

        def new_remote_method(self):
            fixture = self
            def callable_to_call():
                fixture.method_called += 1
                if fixture.exception:
                    raise DomainException('ex')
            return RemoteMethod('amethod', callable_to_call, self.method_result, immutable=False)

        @scenario
        def success(self):
            self.exception = False
            self.expected_response = 'success: method was called 1 times'

        @scenario
        def exception(self):
            self.exception = True
            self.expected_response = 'exception: method was called 1 times'

    @test(RegenerateMethodResultScenarios)
    def regenerating_method_results(self, fixture):
        """If a MethodResult is set up to replay_request=True, the view it is part of (and thus itself) is recreated
           before the (new incarnation of the) MethodResult generates its actual response. Replaying the request means recreating
           all Widgets on the current View as well as the MethodResult itself. The construction of any of these 
           objects may happen differently because of the changes made during the RemoteMethod's execution. Replaying
           the request ensures that the MethodResult reflects such changes, yet ensures that the RemoteMethod
           is not executed twice.
        """

        wsgi_app = fixture.new_wsgi_app(remote_method=fixture.remote_method)
        browser = Browser(wsgi_app)
        
        with CallMonitor(fixture.system_control.orm_control.commit) as monitor:
            browser.post('/_amethod_method', {})
        vassert( browser.raw_html == fixture.expected_response )
        vassert( monitor.times_called == 2 )


    class WidgetResultScenarios(WebFixture):
        changes_made = False
        @stubclass(Widget)
        class WidgetChildStub(object):
            is_Widget = True
            def __init__(self, text):
                self.text = text
            def render(self): return '<%s>' % self.text
            def get_js(self, context=None): return ['js(%s)' % self.text]
            
        def new_WidgetStub(self):
            fixture = self
            @stubclass(Widget)
            class WidgetStub(Widget):
                css_id = 'an_id'
                def __init__(self, view):
                    super(WidgetStub, self).__init__(view)
                    message = 'changed contents' if fixture.changes_made else 'initial contents'
                    self.add_child(fixture.WidgetChildStub(message))
            return WidgetStub

        def new_WidgetWithRemoteMethod(self):
            fixture = self
            @stubclass(Widget)
            class WidgetWithRemoteMethod(Widget):
                def __init__(self, view):
                    super(WidgetWithRemoteMethod, self).__init__(view)
                    method_result = WidgetResult(fixture.WidgetStub(view), as_json_and_result=True)
                    def change_something():
                        fixture.changes_made = True
                        if fixture.exception:
                            raise DomainException('ex')
                    remote_method = RemoteMethod('amethod', change_something, default_result=method_result,
                                                 immutable=False)
                    view.add_resource(remote_method)
            return WidgetWithRemoteMethod

        @scenario
        def success(self):
            self.exception = False
            self.expected_response = {'success': True, 
                                      'widget': '<changed contents><script type="text/javascript">js(changed contents)</script>'}

        @scenario
        def exception(self):
            self.exception = True
            self.expected_response = {'success': False, 
                                      'widget': '<changed contents><script type="text/javascript">js(changed contents)</script>'}

    @test(WidgetResultScenarios)
    def widgets_that_change_during_method_processing(self, fixture):
        """The Widget rendered by WidgetResult reflects its Widget as it would have
           looked if it were constructed AFTER the changes effected by executing 
           its RemoteMethod have been committed.
        """
    
        wsgi_app = fixture.new_wsgi_app(child_factory=fixture.WidgetWithRemoteMethod.factory())
        browser = Browser(wsgi_app)
        
        browser.post('/_amethod_method', {})
        json_response = json.loads(browser.raw_html)
        vassert( json_response == fixture.expected_response )



# Copyright 2013, 2014, 2015 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
from nose.tools import istest
from reahl.tofu import Fixture, test
from reahl.tofu import vassert

from reahl.web_dev.fixtures import WebFixture
from reahl.web.fw import UserInterface, IdentityDictionary, Bookmark
from reahl.web.ui import HTML5Page
from reahl.webdev.tools import Browser
from reahl.component.i18n import Translator


from reahl.web_dev.fixtures import WebBasicsMixin
from reahl.web.fw import UrlBoundView, Url
from reahl.component.context import ExecutionContext
from reahl.webdev.tools import WidgetTester

class XXFixture(Fixture, WebBasicsMixin):
    def new_view(self):
        current_path = Url(ExecutionContext.get_context().request.url).path
        view = UrlBoundView(None, current_path, 'Harness view', {})
        return view

@istest
class XXTests(object):
    @test(XXFixture)
    def test(self, fixture):
        from reahl.web.ui import P

        p = P(fixture.view)
        tester = WidgetTester(p)
        rendered_html = tester.render_html()
        vassert( rendered_html == '<p></p>' )


        
    
@test(WebFixture)
def i18n_urls(fixture):
    """The current locale is determined by reading the first segment of the path. If the locale is not present in the
    path, web.default_url_locale is used."""
    _ = Translator('reahl-web')

    class I18nUI(UserInterface):
        def assemble(self):
            view = self.define_view('/aview', title=_('A View'))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            self.define_user_interface('/a_ui',  I18nUI,  IdentityDictionary(), name='test_ui')
            
    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/a_ui/aview')
    vassert( browser.title == 'A View' )

    browser.open('/af/a_ui/aview')
    vassert( browser.title == '\'n Oogpunt' )

    fixture.context.config.web.default_url_locale = 'af'
    browser.open('/a_ui/aview')
    vassert( browser.title == '\'n Oogpunt' )
    
    browser.open('/en_gb/a_ui/aview')
    vassert( browser.title == 'A View' )


@test(WebFixture)
def bookmarks(fixture):
    """Bookmarks normally refer to the current locale. You can override that to be a specified locale instead.
    """

    bookmark = Bookmark('/base_path', '/relative_path', 'description')
    af_bookmark = Bookmark('/base_path', '/relative_path', 'description', locale='af')

    vassert( af_bookmark.locale == 'af' )
    vassert( af_bookmark.href.path == '/af/base_path/relative_path' )

    vassert( bookmark.locale is None )
    vassert( bookmark.href.path == '/base_path/relative_path' )



# Copyright 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import vassert, scenario, expected, test, NoException, Fixture

from reahl.webdev.tools import WidgetTester
from reahl.web_dev.fixtures import WebFixture


from reahl.component.exceptions import ProgrammerError
from reahl.web.ui import HTMLAttributeValueOption
from reahl.web.bootstrap.ui import Div
from reahl.web.bootstrap.grid import ColumnLayout, ResponsiveSize, Container, DeviceClass


@test(WebFixture)
def containers(fixture):
    """There are two types of Bootstrap containers:  a full width container, and a responsive (fluid) container."""

    widget = Div(fixture.view).use_layout(Container())
    tester = WidgetTester(widget)
    
    css_class = tester.xpath('//div')[0].attrib['class']
    vassert( 'container' == css_class )

    widget = Div(fixture.view).use_layout(Container(fluid=True))
    tester = WidgetTester(widget)
    
    css_class = tester.xpath('//div')[0].attrib['class']
    vassert( 'container-fluid' == css_class )


@test(WebFixture)
def column_layout_basics(fixture):
    """The bootstrap.ColumnLayout adds the correct classes for Bootstrap to lay out its Widget as a row with columns."""

    layout = ColumnLayout(('column_a', ResponsiveSize(lg=4)), ('column_b', ResponsiveSize(lg=8)))
    widget = Div(fixture.view)
    
    vassert( not widget.has_attribute('class') )
    
    widget.use_layout(layout)

    vassert( widget.get_attribute('class') == 'row' )
    column_a, column_b = widget.children

    vassert( 'col-lg-4' in column_a.get_attribute('class')  )
    vassert( 'col-lg-8' in column_b.get_attribute('class')  )


@test(WebFixture)
def column_layout_sizes(fixture):
    """It is mandatory to specify sizes for all columns."""

    with expected(ProgrammerError):
        ColumnLayout('column_a')


@test(WebFixture)
def adding_columns(fixture):
    """You can add additional columns after construction."""

    widget = Div(fixture.view).use_layout(ColumnLayout())

    vassert( not widget.children )

    widget.layout.add_column(ResponsiveSize(lg=4))

    [added_column] = widget.children
    vassert( added_column.get_attribute('class') == 'col-lg-4' )


@test(WebFixture)
def allowed_sizes(fixture):
    """The device classes for which sizes can be specified."""
    size = ResponsiveSize(xs=1, sm=2, md=3, lg=4, xl=5)

    vassert( size == {'xs':1, 'sm':2, 'md':3, 'lg':4, 'xl':5} )


@test(WebFixture)
def column_offsets(fixture):
    """You can optionally specify space to leave empty (an offset) before a column at specific device sizes."""

    layout = ColumnLayout(('column_a', ResponsiveSize(xl=2).offset(xs=2, sm=4, md=6, lg=3, xl=1)))
    widget = Div(fixture.view).use_layout(layout)

    [column_a] = layout.columns.values()

    vassert( 'col-xl-2' in column_a.get_attribute('class')  )
    vassert( 'col-lg-offset-3' in column_a.get_attribute('class')  )
    vassert( 'col-xs-offset-2' in column_a.get_attribute('class')  )
    vassert( 'col-sm-offset-4' in column_a.get_attribute('class')  )
    vassert( 'col-md-offset-6' in column_a.get_attribute('class')  )
    vassert( 'col-xl-offset-1' in column_a.get_attribute('class')  )


@test(WebFixture)
def column_clearfix(fixture):
    """If a logical row spans more than one visual row for a device size, bootstrap clearfixes are
       automatically inserted to ensure cells in resultant visual rows are neatly arranged.
    """

    # Case: Adding a correct clearfix in the right place
    wrapping_layout = ColumnLayout(('column_a', ResponsiveSize(xs=8).offset(xs=2)),
                                   ('column_b', ResponsiveSize(xs=2).offset(xs=2))
    )
    widget = Div(fixture.view).use_layout(wrapping_layout)

    [column_a, clearfix, column_b] = widget.children           
    vassert( [column_a, column_b] == [i for i in wrapping_layout.columns.values()] )
    vassert( 'clearfix' in clearfix.get_attribute('class')  )
    vassert( 'visible-xs-block' in clearfix.get_attribute('class')  )

    # Case: When clearfix needs to take "implicit" sizes of smaller device classes into account
    wrapping_layout = ColumnLayout(('column_a', ResponsiveSize(xs=8).offset(xs=2)),
                                   ('column_b', ResponsiveSize(lg=2).offset(lg=2))
    )
    widget = Div(fixture.view).use_layout(wrapping_layout)

    [column_a, clearfix, column_b] = widget.children           
    vassert( [column_a, column_b] == [i for i in wrapping_layout.columns.values()] )
    vassert( 'clearfix' in clearfix.get_attribute('class')  )
    vassert( 'visible-lg-block' in clearfix.get_attribute('class')  )

    # Case: When no clearfix must be added
    non_wrapping_layout = ColumnLayout(('column_a', ResponsiveSize(xs=2).offset(xs=2)),
                                       ('column_b', ResponsiveSize(xs=2))
    )
    widget = Div(fixture.view).use_layout(non_wrapping_layout)

    [column_a, column_b] = widget.children
    vassert( [column_a, column_b] == [i for i in non_wrapping_layout.columns.values()] )  


@test(Fixture)
def allowed_string_options(fixture):
    """The value of an HTMLAttributeValueOption is constrained to one of its stated valid options if it is set."""
    with expected(NoException):
        HTMLAttributeValueOption('validoption', True, constrain_value_to=['anoption', 'anotheroption', 'validoption'])

    with expected(NoException):
        HTMLAttributeValueOption('invalidoption', False, constrain_value_to=['anoption', 'anotheroption', 'validoption'])

    with expected(ProgrammerError):
        HTMLAttributeValueOption('invalidoption', True, constrain_value_to=['anoption', 'anotheroption', 'validoption'])


@test(Fixture)
def composed_class_string(fixture):
    """A HTMLAttributeValueOption be made into as a css class string."""
    style_class = HTMLAttributeValueOption('validoption', True, prefix='pre', constrain_value_to=['validoption'])
    vassert( style_class.as_html_snippet() == 'pre-validoption' )


@test(WebFixture)
def all_device_classes(fixture):
    """There is a specific list of supported DeviceClasses, in order of device size."""
    device_classes = [ i.class_label for i in DeviceClass.all_classes()]

    vassert( device_classes == ['xs', 'sm', 'md', 'lg', 'xl'] )


@test(WebFixture)
def device_class_identity(fixture):
    """Each supported DeviceClass is identified by a string; you cannot create one that is not supported."""

    device_class = DeviceClass('lg')

    vassert( device_class.class_label == 'lg' )

    def check_ex(ex):
        vassert( six.text_type(ex).startswith('"unsupported" should be one of: "xs","sm","md","lg","xl"') )

    with expected(ProgrammerError, test=check_ex):
        DeviceClass('unsupported')



@test(WebFixture)
def previous_device_class(fixture):
    """You can find the supported DeviceClass smaller than a given one."""

    device_class = DeviceClass('sm')
    vassert( device_class.one_smaller.class_label == 'xs' )

    # Case: when there is no class smaller than given

    device_class = DeviceClass('xs')
    vassert( device_class.one_smaller == None )


@test(WebFixture)
def previous_device_classes(fixture):
    """You can find the ordered list of all supported DeviceClasses smaller than a given one."""

    device_class = DeviceClass('md')

    previous_device_classes = [ i.class_label for i in device_class.all_smaller]
    vassert( previous_device_classes == ['xs', 'sm'] )

    # Case: when there is no class smaller than given
    device_class = DeviceClass('xs')

    previous_device_classes = device_class.all_smaller
    vassert( previous_device_classes == [] )



# Copyright 2013-2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import warnings
import re

from nose.tools import istest
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert, expected
from reahl.stubble import EmptyStub
from reahl.component.py3compat import html_escape

from reahl.web.ui import HTMLElement, PrimitiveInput, TextInput, Form, ValidationException, \
                          CheckboxInput, TextInput, Label, InputLabel, ButtonInput,\
                          PasswordInput, P,\
                          TextArea, SelectInput, RadioButtonInput

from reahl.web.ui import Button, LabelOverInput, CueInput, LabelledInlineInput, LabelledBlockInput

from reahl.component.modelinterface import Field, EmailField, BooleanField,\
                             RequiredConstraint, MinLengthConstraint, PatternConstraint, RemoteConstraint,\
                             Event, Allowed, exposed, Action, Choice, ChoiceGroup, ChoiceField, IntegerField,\
                             MultiChoiceField, DateField
from reahl.component.exceptions import IsInstance
from reahl.web_dev.fixtures import WebFixture
from reahl.webdev.tools import WidgetTester
from reahl.webdev.tools import XPath
from reahl.component_dev.fieldtests import FieldMixin


class InputMixin(FieldMixin):
    def new_form(self):
        return Form(self.view, 'test')

    def new_event(self, label='click me'):
        event = Event(label=label)
        event.bind('aname', None)
        self.form.define_event_handler(event)
        return event


class SimpleInputFixture(WebFixture, InputMixin):
    pass


class InputMixin2(InputMixin):
    def new_field(self, label='the label'):
        return Field(label=label)
    def new_model_object(self):
        fixture = self
        class ModelObject(object):
            def handle_event(self):
                pass
            @exposed
            def events(self, events):
                events.an_event = Event(label='click me', action=Action(self.handle_event))
            @exposed
            def fields(self, fields):
                fields.an_attribute = fixture.field
        return ModelObject()


class InputStateFixture(WebFixture, InputMixin):
    def new_field(self):
        field = EmailField(default='default value')
        field.bind('an_attribute', self.model_object)
        return field

    def new_input(self):    
        return PrimitiveInput(self.form, self.field)

    @scenario
    def defaulted_input_no_value_on_domain(self):
        del self.model_object.an_attribute
        self.input.prepare_input()

        self.expected_state = 'defaulted'
        self.expected_value = 'default value'

    @scenario
    def defaulted_input_value_on_domain(self):
        self.input.prepare_input()
        self.model_object.an_attribute = 'something'

        self.expected_state = 'defaulted'
        self.expected_value = 'something'

    @scenario
    def validly_entered(self):
        self.input.enter_value('someone@home.org')
        self.input.prepare_input()

        self.expected_state = 'validly_entered' 
        self.expected_value = 'someone@home.org'

    @scenario
    def empty_input_non_required_field(self):
        self.input.enter_value('')
        self.input.prepare_input()

        self.expected_state = 'validly_entered' 
        self.expected_value = ''

    @scenario
    def empty_input_required_field(self):
        self.field.make_required('')
        self.input.enter_value('')
        self.input.prepare_input()

        self.expected_state = 'invalidly_entered' 
        self.expected_value = ''

    @scenario
    def invalidly_entered(self):
        self.input.enter_value('not an email address')
        self.input.prepare_input()

        self.expected_state = 'invalidly_entered' 
        self.expected_value = 'not an email address'


@test(InputStateFixture)
def the_states_of_an_input(fixture):
    """An Input can be in one of three states: defaulted, invalidly_entered or validly_entered. Depending
       on the state of the input, it will have a different value when rendering.

       The states:
         - defaulted: the user has not entered anything
           value:     the value on the model object which is bound to its Field if that object
                      has such an an_attribute, else the default value of the Field.
         - invalidly_entered: the user entered a value which failed validation
           value:             the invalid value which the user entered
         - validly_entered: the user entered a valid value
           value:           the value which the user entered even if the underlying
                            model object has a corresponding value
    """
    vassert( fixture.input.get_input_status() == fixture.expected_state )
    vassert( fixture.input.value == fixture.expected_value )


class Scenarios(WebFixture, InputMixin):
    @scenario
    def text_input(self):
        self.widget = TextInput(self.form, self.field)
        self.expected_html = r'<input name="an_attribute" form="test" type="text" value="field value" class="reahl-textinput">'
        self.field_controls_visibility = True

    @scenario
    def text_input_placeholder_default(self):
        self.widget = TextInput(self.form, self.field, placeholder=True)
        self.expected_html = r'<input name="an_attribute" form="test" placeholder="the label" type="text" value="field value" class="reahl-textinput">'
        self.field_controls_visibility = True

    @scenario
    def text_input_placeholder_specified(self):
        self.widget = TextInput(self.form, self.field, placeholder="some text")
        self.expected_html = r'<input name="an_attribute" form="test" placeholder="some text" type="text" value="field value" class="reahl-textinput">'
        self.field_controls_visibility = True

    @scenario
    def input_label_deprecated_interface(self):
        html_input = TextInput(self.form, self.field)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            self.widget = InputLabel(html_input)
        self.expected_html = r'<label for="%s">the label</label>' % html_input.css_id
        self.field_controls_visibility = True

    @scenario
    def input_label(self):
        html_input = TextInput(self.form, self.field)
        self.widget = Label(self.view, for_input=html_input)
        self.expected_html = r'<label for="%s">the label</label>' % html_input.css_id
        self.field_controls_visibility = True

    @scenario
    def input_label_with_text(self):
        html_input = TextInput(self.form, self.field)
        self.widget = Label(self.view, for_input=html_input, text='some text')
        self.expected_html = r'<label for="%s">some text</label>' % html_input.css_id
        self.field_controls_visibility = True

    @scenario
    def button_input(self):
        self.widget = self.form.add_child(ButtonInput(self.form, self.event))
        self.expected_html = r'<input name="event.aname\?" form="test" type="submit" value="click me">'
        self.field_controls_visibility = False

    @scenario
    def button(self):
        self.widget = self.form.add_child(Button(self.form, self.event))
        self.expected_html = r'<span class="reahl-button"><input name="event\.aname\?" form="test" type="submit" value="click me"></span>'
        self.field_controls_visibility = False

    @scenario
    def labelled_inline_input(self):
        self.model_object.an_attribute = 'field value'
        self.widget = self.form.add_child(LabelledInlineInput(TextInput(self.form, self.field)))
        self.expected_html = r'<span class="reahl-labelledinput">'\
                             r'<label for="(.*)">the label</label>' \
                             r'<span><input name="an_attribute" id="\1" form="test" type="text" value="field value" class="reahl-textinput"></span>' \
                             r'</span>'
        self.field_controls_visibility = True

    @scenario
    def labelled_block_input(self):
        self.model_object.an_attribute = 'field value'
        self.widget = self.form.add_child(LabelledBlockInput(TextInput(self.form, self.field)))
        self.expected_html = r'<div class="pure-g reahl-labelledinput">'\
                             r'<div class="column-label pure-u-1-4"><label for="(.*)">the label</label></div>' \
                             r'<div class="column-input pure-u-3-4"><input name="an_attribute" id="\1" form="test" type="text" value="field value" class="reahl-textinput"></div>' \
                             r'</div>'
        self.field_controls_visibility = True

    @scenario
    def password(self):
        self.widget = self.form.add_child(PasswordInput(self.form, self.field))
        self.expected_html = r'<input name="an_attribute" form="test" type="password">'
        self.field_controls_visibility = True


    def setup_checkbox_scenario(self, boolean_value):
        self.model_object.an_attribute = boolean_value

        self.field = BooleanField(required=True, label='my text', required_message='$label is needed here')
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(CheckboxInput(self.form, self.field))
        self.field_controls_visibility = True

    @scenario
    def checkbox_true(self):
        """A checkbox needs a 'checked' an_attribute if its field is True. It also renders ONLY the
            validation message for its required validation_constraint, not for all constraints"""
        self.setup_checkbox_scenario(True)
        self.expected_html = r'<input name="an_attribute" checked="checked" form="test" required="\*" type="checkbox" ' \
                             r'class="{&quot;validate&quot;: {&quot;messages&quot;: {&quot;required&quot;: &quot;my text is needed here&quot;}}}">'

    @scenario
    def checkbox_false(self):
        self.setup_checkbox_scenario(False)
        self.expected_html = r'<input name="an_attribute" form="test" required="\*" type="checkbox" '\
                             r'class="{&quot;validate&quot;: {&quot;messages&quot;: {&quot;required&quot;: &quot;my text is needed here&quot;}}}">'

    @scenario
    def text_area_input(self):
        self.widget = self.form.add_child(TextArea(self.form, self.field, rows=30, columns=20))
        self.expected_html = r'<textarea name="an_attribute" cols="20" rows="30">field value</textarea>'
        self.field_controls_visibility = True

    @scenario
    def select_input(self):
        self.model_object.an_attribute = 2

        choices = [Choice(False, BooleanField(label='None')),
                   ChoiceGroup('grouped', [
                       Choice(1, IntegerField(label='One')),
                       Choice(2, IntegerField(label='Two'))])
                  ]
        self.field = ChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(SelectInput(self.form, self.field))
        group = r'<optgroup label="grouped"><option value="1">One</option><option selected="selected" value="2">Two</option></optgroup>'
        option = r'<option value="off">None</option>'
        self.expected_html = r'<select name="an_attribute" form="test">%s%s</select>' % (option, group)
        self.field_controls_visibility = True

    @scenario
    def select_input_multi(self):
        self.model_object.an_attribute = [2]

        choices = [Choice(1, IntegerField(label='One')),
                   Choice(2, IntegerField(label='Two'))
                  ]
        self.field = MultiChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(SelectInput(self.form, self.field))
        options = r'<option value="1">One</option><option selected="selected" value="2">Two</option>'
        self.expected_html = r'<select name="an_attribute" form="test" multiple="multiple">%s</select>' % (options)
        self.field_controls_visibility = True

    @scenario
    def radio_button(self):
        self.model_object.an_attribute = 2

        choices = [ Choice(1, IntegerField(label='One')),
                    Choice(2, IntegerField(label='Two')) ]
        self.field = ChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(RadioButtonInput(self.form, self.field))

        validations = html_escape('{"validate": {"messages": {"pattern": "an_attribute should be one of the following: 1\|2"}}}')
        radio_button = r'<span class="reahl-radio-button">'\
                       r'<input name="an_attribute"%s form="test" pattern="\(1\|2\)" '\
                              r'title="an_attribute should be one of the following: 1\|2" '\
                              r'type="radio" value="%s" class="%s">%s'\
                       r'</span>'

        outer_div = r'<div class="reahl-radio-button-input">%s</div>'
        buttons = (radio_button % ('', '1', validations, 'One')) +\
                  (radio_button % (' checked="checked"', '2', validations, 'Two'))
        self.expected_html = outer_div % buttons
        self.field_controls_visibility = True


@test(Scenarios)
def basic_rendering(fixture):
    """What the rendered html for a number of simple inputs look like."""

    tester = WidgetTester(fixture.widget)
    actual = tester.render_html()
    vassert( re.match(fixture.expected_html, actual) )


@test(Scenarios)
def rendering_when_not_allowed(fixture):
    """When not allowed to see the Widget, it is not rendered."""
    tester = WidgetTester(fixture.widget)

    fixture.field.access_rights.readable = Allowed(False)
    fixture.field.access_rights.writable = Allowed(False)
    actual = tester.render_html()
    if fixture.field_controls_visibility:
        vassert( actual == '' )
    else:
        vassert( re.match(fixture.expected_html, actual) )


@test(SimpleInputFixture)
def backwards_compatibility(fixture):
    """We still support overriding create_html_input for backwards compatibility."""

    # Case when old interface is overridden
    class OldInput(TextInput):
        def create_html_input(self):
            return super(OldInput, self).create_html_widget()

    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter('always')
        html_input = OldInput(fixture.form, fixture.field)

    vassert( caught_warnings )
    tester = WidgetTester(html_input)

    actual = tester.render_html()
    vassert( actual ) # Something was rendered.

    # Case when new interface is used
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter('always')
        html_input = TextInput(fixture.form, fixture.field)
    vassert( not caught_warnings )


@test(SimpleInputFixture)
def input_wrapped_widgets(fixture):
    """An Input is an empty Widget; its contents are supplied by overriding its 
       .create_html_widget() method. Several methods for setting HTML-things, like 
       css id are delegated to this Widget which represents the Input in HTML.
    """
    class MyInput(PrimitiveInput):
        def create_html_widget(self):
            return HTMLElement(self.view, 'x')

    test_input = MyInput(fixture.form, fixture.field)
    tester = WidgetTester(test_input)

    rendered = tester.render_html()
    vassert( rendered == '<x>' )

    test_input.set_id('myid')
    test_input.set_title('mytitle')
    test_input.add_to_attribute('list-attribute', ['one', 'two'])
    test_input.set_attribute('an-attribute', 'a value')

    rendered = tester.render_html()
    vassert( rendered == '<x id="myid" an-attribute="a value" list-attribute="one two" title="mytitle">' )


@test(SimpleInputFixture)
def wrong_args_to_input(fixture):
    """Passing the wrong arguments upon constructing an Input results in an error."""

    with expected(IsInstance):
        PrimitiveInput(fixture.form, EmptyStub())

    with expected(IsInstance):
        PrimitiveInput(EmptyStub(), Field())


class CheckboxFixture(WebFixture, InputMixin2):
    def new_field(self):
        return BooleanField(label='my text')


@test(CheckboxFixture)
def marshalling_of_checkbox(fixture):
    """When a form is submitted, the value of a checkbox is derived from
       whether the checkbox is included in the submission or not."""

    model_object = fixture.model_object
    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            checkbox = self.add_child(CheckboxInput(self, model_object.fields.an_attribute))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(Button(self, model_object.events.an_event))
            fixture.checkbox = checkbox

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'))
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')

    # Case: checkbox is submitted with form (ie checked)
    fixture.driver_browser.check("//input[@type='checkbox']")
    fixture.driver_browser.click("//input[@value='click me']")

    vassert( model_object.an_attribute )
    vassert( fixture.checkbox.value == 'on' )

    # Case: checkbox is not submitted with form (ie unchecked)
    fixture.driver_browser.uncheck("//input[@type='checkbox']")
    fixture.driver_browser.click("//input[@value='click me']")

    vassert( not model_object.an_attribute )
    vassert( fixture.checkbox.value == 'off' )


class LabelOverInputFixture(WebFixture, InputMixin2):
    label_xpath = "//form/span[@class='reahl-labelledinput reahl-labeloverinput']/label[text()='the label']"


@test(LabelOverInputFixture)
def label_behaviour(fixture):
    """The label is displayed over the input, but it disappears when selected
       or when there's a value in the input."""

    model_object = fixture.model_object

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(LabelOverInput(TextInput(self, model_object.fields.an_attribute)))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(Button(self, model_object.events.an_event))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    fixture.reahl_server.set_app(wsgi_app)

    # Case: the field is empty and does not have a default value
    # - render initially with label
    fixture.driver_browser.open('/')
    vassert( fixture.driver_browser.is_element_present(fixture.label_xpath) )
    fixture.driver_browser.wait_for_element_visible(fixture.label_xpath)

    # - if click, label disappears, input gets focus
    fixture.driver_browser.click(fixture.label_xpath)
    fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)

    # - if focus lost, label appears again if nothing is entered
    fixture.driver_browser.press_tab('//input')
    fixture.driver_browser.wait_for_element_visible(fixture.label_xpath)

    # - if focus lost, label does not appear again if something is entered
    fixture.driver_browser.type('//input[@type="text"]', 'my@email.com')
    fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)
    fixture.driver_browser.press_tab('//input')
    fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)


    # Case: the field has a defaulted value
    model_object.an_attribute = 'some value'
    fixture.driver_browser.open('/')

    # - render initially without a label
    fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)

    # - if value removed, and focus lost, label appears
    fixture.driver_browser.type('//input[@type="text"]', '')
    fixture.driver_browser.press_tab('//input')
    fixture.driver_browser.wait_for_element_visible(fixture.label_xpath)


class BasicRenderingLabelOverInputScenarios(LabelOverInputFixture):
    @scenario
    def domain_has_value(self):
        #  - has CSS class reahl-labeloverinput
        #  - the label is hidden
        self.model_object.an_attribute = 'my value'
        self.expected_html = r'<span class="reahl-labelledinput reahl-labeloverinput">'\
               r'<label for="(.*)" hidden="true">the label</label>' \
               r'<span><input name="an_attribute" id="\1" form="test" type="text" value="my value" class="reahl-textinput"></span>'\
               r'</span>'

    @scenario
    def domain_has_no_value(self):
        #  - has CSS class reahl-labeloverinput
        #  - the label is not hidden
        vassert( not hasattr(self.model_object, 'an_attribute') )
        self.expected_html = r'<span class="reahl-labelledinput reahl-labeloverinput">'\
               r'<label for="(.*)">the label</label>' \
               r'<span><input name="an_attribute" id="\1" form="test" type="text" value="" class="reahl-textinput"></span>'\
               r'</span>'

    @scenario
    def not_allowed_to_be_seen(self):
        # - the underlying input is not visible
        self.model_object.an_attribute = 'my value'
        self.field.access_rights.readable = Allowed(False)
        self.field.access_rights.writable = Allowed(False)
        self.expected_html = ''



@test(BasicRenderingLabelOverInputScenarios)
def basic_rendering_label_over(fixture):
    """What the html for a LabelOverInput looks like."""

    fixture.field.bind('an_attribute', fixture.model_object)
    label_over_input = fixture.form.add_child(LabelOverInput(TextInput(fixture.form, fixture.field)))
    tester = WidgetTester(label_over_input)

    actual = tester.render_html()
    vassert( re.match(fixture.expected_html, actual) )



class CueInputFixture(WebFixture, InputMixin2):
    cue_xpath = "//div[contains(@class,'reahl-cueinput')]/div[contains(@class,'reahl-cue')]/p"

    def new_text_input(self):
        return TextInput(self.form, self.field)

    def new_cue(self):
        return P(self.view, text='this is your cue')

    def new_cue_input(self):
        return CueInput(self.text_input, self.cue)

        
@test(CueInputFixture)
def cue_behaviour(fixture):
    """The Cue appears when its input has focus only."""
    model_object = fixture.model_object

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(CueInput(TextInput(self, model_object.fields.an_attribute), P(view, text='this is your cue')))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(Button(self, model_object.events.an_event))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')

    # - initially rendered without cue
    vassert( fixture.driver_browser.is_element_present(fixture.cue_xpath) )
    fixture.driver_browser.wait_for_element_not_visible(fixture.cue_xpath)

    # - if click, cue appears
    fixture.driver_browser.click('//input[@type="text"]')
    fixture.driver_browser.wait_for_element_visible(fixture.cue_xpath)

    # - if focus lost, cue disappears again
    fixture.driver_browser.press_tab('//input')
    fixture.driver_browser.wait_for_element_not_visible(fixture.cue_xpath)


class BasicRenderingCueInputScenarios(CueInputFixture):
    @scenario
    def normal(self):
        self.expected_html = r'<div class="pure-g reahl-cueinput reahl-labelledinput">'\
                r'<div class="column-label pure-u-1-4"><label for="(.*)">the label</label></div>' \
                r'<div class="column-input pure-u-1-2"><input name="an_attribute" id="\1" form="test" type="text" value="my value" class="reahl-textinput"></div>' \
                 r'<div class="column-cue pure-u-1-4 reahl-cue">' \
                 r'<p hidden="true">this is your cue</p>' \
                 r'</div>' \
               r'</div>'

    @scenario
    def not_allowed_to_be_seen(self):
        self.field.access_rights.readable = Allowed(False)
        self.field.access_rights.writable = Allowed(False)
        self.expected_html = ''


@test(BasicRenderingCueInputScenarios)
def rendering_cue_input(fixture):
    """What the html for a CueInput looks like."""
    fixture.model_object.an_attribute = 'my value'
    fixture.field.bind('an_attribute', fixture.model_object)
    cue = fixture.cue
    cue_input = fixture.cue_input
    fixture.form.add_child(cue_input)

    tester = WidgetTester(cue_input)

    actual = tester.render_html()
    vassert( re.match(fixture.expected_html, actual) )



class FuzzyTextInputFixture(WebFixture, InputMixin2):
    def new_field(self, label='the label'):
        return DateField(label=label)


@test(FuzzyTextInputFixture)
def fuzzy(fixture):
    """A TextInput can be created as fuzzy=True. Doing this results in the possibly imprecise
       input that was typed by the user to be interpreted server-side and changed to the
       more exact representation in the client browser.  This happens upon the input losing focus."""
    model_object = fixture.model_object
    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(LabelledBlockInput(TextInput(self, model_object.fields.an_attribute, fuzzy=True)))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(Button(self, model_object.events.an_event))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    fixture.reahl_server.set_app(wsgi_app)
    browser = fixture.driver_browser
    browser.open('/')

    browser.type(XPath.input_labelled('the label'), '20 November 2012')
    browser.press_tab(XPath.input_labelled('the label'))
    browser.wait_for(browser.is_element_value, XPath.input_labelled('the label'), '20 Nov 2012')








# Copyright 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import vassert, scenario, expected, test

from reahl.webdev.tools import XPath
from reahl.web_dev.fixtures import WebFixture

from reahl.component.modelinterface import exposed, BooleanField

from reahl.web.bootstrap.ui import Div, P
from reahl.web.bootstrap.forms import Form, FormLayout, CheckboxInput
from reahl.web.bootstrap.popups import PopupA, CheckCheckboxScript

class PopupAFixture(WebFixture):
    # (note that this xpath ensures that the p is the ONLY content of the dialog)
    poppedup_contents = "//div[@class='modal-body' and count(*)=1]/p[@id='contents']"

    def new_wsgi_app(self):
        return super(PopupAFixture, self).new_wsgi_app(enable_js=True,
                                                       child_factory=self.MainWidget.factory())
    def new_webconfig(self):
        webconfig = super(PopupAFixture, self).new_webconfig()
        webconfig.frontend_libraries.enable_experimental_bootstrap()
        return webconfig


class PopupAFixtureWithContent(PopupAFixture):
    def new_MainWidget(self):
        fixture = self

        class PopupTestPanel(Div):
            def __init__(self, view):
                super(PopupTestPanel, self).__init__(view)
                self.add_child(PopupA(view, view.as_bookmark(), '#contents'))
                popup_contents = self.add_child(P(view, text='this is the content of the popup'))
                popup_contents.set_id('contents')

        return PopupTestPanel

@test(PopupAFixtureWithContent)
def default_behaviour(fixture):
    """If you click on the A, a popupwindow opens with its contents the specified
       element on the target page."""

    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser
    browser.open('/')

    # The A is rendered correctly
    browser.is_element_present("//a[@title='Home page' and text()='Home page' and @href='/']")

    # subsequent behaviour
    browser.click(XPath.link_with_text('Home page'))
    browser.wait_for_element_visible(fixture.poppedup_contents)

    browser.click(XPath.button_labelled('Close'))
    browser.wait_for_element_not_visible(fixture.poppedup_contents)


class PopupAFixtureWithCustomButtons(PopupAFixture):
    def new_MainWidget(self):
        fixture = self

        class PopupTestPanel(Div):
            def __init__(self, view):
                super(PopupTestPanel, self).__init__(view)
                popup_a = self.add_child(PopupA(view, view.as_bookmark(), '#contents'))
                popup_a.add_js_button('Butt1')
                popup_a.add_js_button('Butt2')
                popup_contents = self.add_child(P(view, text='this is the content of the popup'))
                popup_contents.set_id('contents')

        return PopupTestPanel


@test(PopupAFixtureWithCustomButtons)
def customising_dialog_buttons(fixture):
    """The buttons of the dialog can be customised."""

    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser
    browser.open('/')

    button1_xpath = XPath.button_labelled('Butt1')
    button2_xpath = XPath.button_labelled('Butt2')
    fixture.driver_browser.open('/')

    browser.click(XPath.link_with_text('Home page'))
    browser.wait_for_element_visible(fixture.poppedup_contents)

    vassert( browser.is_element_present(button1_xpath) )
    vassert( browser.is_element_present(button2_xpath) )


class PopupAFixtureWithCheckBox(PopupAFixture):
    def new_MainWidget(self):
        fixture = self

        class PopupTestPanel(Div):
            @exposed
            def fields(self, fields):
                fields.field = BooleanField(label='a checkbox')

            def __init__(self, view):
                super(PopupTestPanel, self).__init__(view)
                popup_a = self.add_child(PopupA(view, view.as_bookmark(), '#contents'))
                popup_contents = self.add_child(P(view, text='this is the content of the popup'))
                popup_contents.set_id('contents')
                form = self.add_child(Form(view, 'aform')).use_layout(FormLayout())
                checkbox = form.layout.add_input(CheckboxInput(form, self.fields.field))

                popup_a.add_js_button('Checkit', CheckCheckboxScript(checkbox))

        return PopupTestPanel


@test(PopupAFixtureWithCheckBox)
def workings_of_check_checkbox_button(fixture):
    """A CheckCheckBoxButton checks the checkbox on the original page when clicked."""

    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser
    browser.open('/')

    browser.click(XPath.link_with_text('Home page'))
    browser.wait_for_element_visible(fixture.poppedup_contents)

    browser.click(XPath.button_labelled('Checkit'))
    browser.wait_for_element_not_visible(fixture.poppedup_contents)

    vassert( fixture.driver_browser.is_checked(XPath.input_labelled('a checkbox')) )
